//
//  EXFaceDetectorUtils.m
//  Exponent
//
//  Created by Stanisław Chmiela on 22.11.2017.
//  Copyright © 2017 650 Industries. All rights reserved.
//

#import <EXFaceDetector/EXFaceDetectorUtils.h>
#import <EXFaceDetector/EXCSBufferOrientationCalculator.h>
#import <Firebase/Firebase.h>

NSString *const EXGMVDataOutputWidthKey = @"Width";
NSString *const EXGMVDataOutputHeightKey = @"Height";

static const NSString *kModeOptionName = @"mode";
static const NSString *kDetectLandmarksOptionName = @"detectLandmarks";
static const NSString *kRunClassificationsOptionName = @"runClassifications";
static const NSString *kTrackingEnabled = @"tracking";

@implementation EXFaceDetectorUtils

+ (NSDictionary *)constantsToExport
{
  return @{
           @"Mode" : @{
               @"fast" : @(FIRVisionFaceDetectorPerformanceModeFast),
               @"accurate" : @(FIRVisionFaceDetectorPerformanceModeAccurate)
               },
           @"Landmarks" : @{
               @"all" : @(FIRVisionFaceDetectorLandmarkModeAll),
               @"none" : @(FIRVisionFaceDetectorLandmarkModeNone)
               },
           @"Classifications" : @{
               @"all" : @(FIRVisionFaceDetectorClassificationModeAll),
               @"none" : @(FIRVisionFaceDetectorClassificationModeNone)
               }
           };
}

+ (BOOL)didOptionsChange:(FIRVisionFaceDetectorOptions *)options comparingTo:(FIRVisionFaceDetectorOptions *)other
{
  return options.performanceMode == other.performanceMode &&
  options.classificationMode == other.classificationMode &&
  options.contourMode == other.contourMode &&
  options.minFaceSize == other.minFaceSize &&
  options.landmarkMode == other.landmarkMode &&
  options.trackingEnabled == other.trackingEnabled;
}

+ (FIRVisionFaceDetectorOptions *)createCopy:(FIRVisionFaceDetectorOptions *)from
{
  FIRVisionFaceDetectorOptions *options = [FIRVisionFaceDetectorOptions new];
  options.performanceMode = from.performanceMode;
  options.classificationMode = from.classificationMode;
  options.contourMode = from.contourMode;
  options.minFaceSize = from.minFaceSize;
  options.landmarkMode = from.landmarkMode;
  options.trackingEnabled = from.trackingEnabled;
  return options;
}


+ (FIRVisionFaceDetectorOptions *) mapOptions:(NSDictionary*)options {
  return [self newOptions:[FIRVisionFaceDetectorOptions new] withValues:options];
}

+ (FIRVisionFaceDetectorOptions *) newOptions:(FIRVisionFaceDetectorOptions*)options withValues:(NSDictionary *)values
{
  FIRVisionFaceDetectorOptions *result = [self createCopy:options];
  if([values objectForKey:kModeOptionName]) {
    result.performanceMode = [values[kModeOptionName] longValue];
  }
  if([values objectForKey:kDetectLandmarksOptionName]) {
    result.landmarkMode = [values[kDetectLandmarksOptionName] longValue];
  }
  if([values objectForKey:kRunClassificationsOptionName]) {
    result.classificationMode = [values[kRunClassificationsOptionName] longValue];
  }
  if([values objectForKey:kTrackingEnabled]) {
    result.trackingEnabled = [values[kTrackingEnabled] boolValue];
  }
  return result;
}

+ (BOOL) areOptionsEqual:(FIRVisionFaceDetectorOptions *)first to:(FIRVisionFaceDetectorOptions*)second {
  return [self didOptionsChange:first comparingTo:second];
}

+ (NSDictionary *)defaultFaceDetectorOptions
{
  return @{
           kModeOptionName: @(FIRVisionFaceDetectorPerformanceModeFast),
           kDetectLandmarksOptionName: @(FIRVisionFaceDetectorLandmarkModeNone),
           kRunClassificationsOptionName: @(FIRVisionFaceDetectorClassificationModeNone)
           };
}

+ (int)toCGImageOrientation:(UIImageOrientation)imageOrientation
{
  switch (imageOrientation) {
    case UIImageOrientationUp:
      return kCGImagePropertyOrientationUp;
      break;
    case UIImageOrientationUpMirrored:
      return kCGImagePropertyOrientationUpMirrored;
      break;
    case UIImageOrientationDown:
      return kCGImagePropertyOrientationDown;
      break;
    case UIImageOrientationDownMirrored:
      return kCGImagePropertyOrientationDownMirrored;
      break;
    case UIImageOrientationRight:
      return kCGImagePropertyOrientationRight;
      break;
    case UIImageOrientationRightMirrored:
      return kCGImagePropertyOrientationRightMirrored;
      break;
    case UIImageOrientationLeft:
      return kCGImagePropertyOrientationLeft;
      break;
    case UIImageOrientationLeftMirrored:
      return kCGImagePropertyOrientationLeftMirrored;
      break;
  }
};

# pragma mark - Encoder helpers

+ (EXFaceDetectionAngleTransformBlock)angleTransformerFromTransform:(CGAffineTransform)transform
{
  return ^(float angle) {
    return (float)(angle - (atan2(transform.b, transform.a) * (180 / M_PI)));
  };
}

@end
